-- Lua Log: %LOCALAPPDATA%\Firaxis Games\Sid Meier's Civilization VI\Logs\Lua.log
-- RANGER_IGNORE_FOREST_MOVEMENT_PENALTY
-- ALPINE_IGNORE_HILLS_MOVEMENT_PENALTY

local MODIFIER_ID_SPREAD_DISTANCE = "AI_PREACHERS_SPREAD_DISTANCE"
local MODIFIER_ID_SPREAD_STRENGTH = "AI_SPREAD_STRENGTH"

-- Function to check if a plot is visible to any human player
function IsVisibleToAnyHuman(pPlot)
    for _, pPlayer in ipairs(PlayerManager.GetAliveMajors()) do
        if pPlayer:IsHuman() then
            local visibility = PlayersVisibility[pPlayer:GetID()]:GetState(pPlot:GetIndex())
            if visibility == RevealedState.VISIBLE then
                return true
            end
        end
    end
    return false
end

-- Returns the distance to the closest visible plot if any are visible, otherwise returns -1
function IsUnitAndAdjacentPlotsVisibleToHuman(pUnit, range)
    if not pUnit then
        return -1
    end
    local unitX, unitY = pUnit:GetX(), pUnit:GetY()
    local pPlot = Map.GetPlot(unitX, unitY)
    if not pPlot then
        return -1
    end

    local mapWidth, mapHeight = Map.GetGridSize()
    local minDistance = nil

    -- Iterate over plots within the specified range
    for dx = -range, range do
        for dy = -range, range do
            local targetX = unitX + dx
            local targetY = unitY + dy

            -- Handle map wrapping or bounds
            local validX = true
            local validY = true

            if Map.IsWrapX() then
                targetX = (targetX + mapWidth) % mapWidth
            elseif targetX < 0 or targetX >= mapWidth then
                validX = false
            end

            if Map.IsWrapY() then
                targetY = (targetY + mapHeight) % mapHeight
            elseif targetY < 0 or targetY >= mapHeight then
                validY = false
            end

            if validX and validY then
                local distance = Map.GetPlotDistance(unitX, unitY, targetX, targetY)
                if distance and distance <= range then
                    local targetPlot = Map.GetPlot(targetX, targetY)
                    if targetPlot and IsVisibleToAnyHuman(targetPlot) then
                        if not minDistance or distance < minDistance then
                            minDistance = distance
                        end
                    end
                end
            end
        end
    end

    if minDistance then
        return minDistance
    else
        return -1
    end
end


-- Ability Type to be added/removed
local ABILITY_TYPE = "ABILITY_AI_HIDDEN_MOVEMENT_BONUS"

-- Function to apply modifiers to AI players
function OnPlayerTurnActivated(playerID)
    pPlayer = Players[playerID]
    print("check player: " .. playerID)
    if pPlayer and pPlayer:IsAlive() and pPlayer:IsMajor() and not pPlayer:IsHuman() then -- reminder NOT HUMAN!
        for _, pUnit in pPlayer:GetUnits():Members() do
            -- possible to grant promotions on modolu id
            -- player has attach modifier by id
            
            -- if pUnit:GetCombat() ~= nil then
            --     pUnitExp = pUnit:GetExperience()
            --     prom = GameInfo.UnitPromotions["PROMOTION_SENTRY"].Index
            --     if not pUnitExp:HasPromotion(prom) then
            --         pUnitExp:SetPromotion(prom)
            --         print("Unit: " .. pUnit:GetName() .. "(" .. pUnit:GetX() .. "," .. pUnit:GetY() .. ") set promotion id: " .. prom)
            --     else
            --         print("Unit: " .. pUnit:GetName() .. "(" .. pUnit:GetX() .. "," .. pUnit:GetY() .. ") has already promotion id: " .. prom)
            --     end
            -- end
            
            distance = IsUnitAndAdjacentPlotsVisibleToHuman(pUnit, 5)
            before = pUnit:GetExtraMoves()
            visibleBefore = -1
            if pUnit:GetProperty("last_visible_in_turn") ~= nil then
                visibleBefore = Game.GetCurrentGameTurn() - pUnit:GetProperty("last_visible_in_turn")
            end
            movesShould = 0
            -- print("Unit: " .. pUnit:GetName() .. "(" .. pUnit:GetX() .. "," .. pUnit:GetY() .. ") distance to visiblity is: " .. distance)
            if distance == -1 then
                movesShould = 5
            else
                movesShould = distance
            end

            if visibleBefore <= 1 and visibleBefore ~= -1 then
                movesShould = math.min(movesShould, visibleBefore + 1)
            end

            -- Always at least 1 bonus movement
            movesShould = movesShould + 1

            -- Init units vdmove property
            if pUnit:GetProperty("vdmove") == nil then
                pUnit:SetProperty("vdmove", 0)
                print("Unit: " .. pUnit:GetName() .. "(" .. pUnit:GetX() .. "," .. pUnit:GetY() .. ")[distance=" .. distance .. ", lastseen=" .. visibleBefore .. "] vdmove init")
            end

            if Game.GetCurrentGameTurn() >= 21 then
                vdmove = pUnit:GetProperty("vdmove")
                if vdmove ~= movesShould then
                    calcMove = movesShould - vdmove
                    pUnit:ChangeExtraMoves(calcMove)
                    print("Unit: " .. pUnit:GetName() .. "(" .. pUnit:GetX() .. "," .. pUnit:GetY() .. ")[distance=" .. distance .. ", lastseen=" .. visibleBefore .. "] set from '" .. before .. "' to '" .. pUnit:GetExtraMoves() .. "' ('" .. calcMove .."')")
                    pUnit:SetProperty("vdmove", movesShould)
                end
            end

            -- Get the unit's position
            local x, y = pUnit:GetX(), pUnit:GetY()
            for dx = -2, 2 do
                for dy = -2, 2 do
                    local targetX = x + dx
                    local targetY = y + dy
                    local pPlot = Map.GetPlot(targetX, targetY)
                    if pPlot then
                        -- Check for barbarian units
                        if pPlot:GetUnitCount() >= 1 then
                            for _, pPlotUnit in pairs(Units.GetUnitsInPlot(pPlot)) do
                                if pPlotUnit:GetOwner() == 63 then
                                    barbDistance = IsUnitAndAdjacentPlotsVisibleToHuman(pPlotUnit, 5)
                                    if barbDistance == -1 then
                                        print("Barbarian " .. pPlotUnit:GetName() .. " (distance=" .. barbDistance .. ") killed at " .. targetX .. ":" .. targetY .. ". +100 gold for player " .. pPlayer:GetID())
                                        UnitManager.Kill(pPlotUnit)
                                        local playerTreasury = pPlayer:GetTreasury()
                                        playerTreasury:ChangeGoldBalance(100)

                                        -- Also check for camp, but only if unit was there
                                        local improvementType = pPlot:GetImprovementType()
                                        if improvementType ~= -1 then
                                            local improvementInfo = GameInfo.Improvements[improvementType]
                                            if improvementInfo and improvementInfo.ImprovementType == "IMPROVEMENT_BARBARIAN_CAMP" then
                                                print("Barbarian camp (distance=" .. barbDistance .. ") destroyed at " .. targetX .. ":" .. targetY)
                                                ImprovementBuilder.SetImprovementType(pPlot, -1)  -- Remove the improvement
                                            end
                                        end
                                    else
                                        print("Barbarian " .. pPlotUnit:GetName() .. " (distance=" .. barbDistance .. ") ignored at " .. targetX .. ":" .. targetY)
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    end
end

-- Wrapper function to handle both district and building constructions
function OnConstructionCompleted(playerID, typeID, plotX, plotY, isBuilding)
    local player = Players[playerID]
    -- Check if the player is human or AI
    if player:IsHuman() or IsVisibleToAnyHuman(Map.GetPlot(plotX, plotY)) then
        return  -- Exit if the player is not AI
    end

    local playerUnits = player:GetUnits()
    local playerReligion = player:GetReligion()
    local religionTypeCreated = playerReligion:GetReligionTypeCreated()
    local plot = Map.GetPlot(plotX, plotY)

    if isBuilding then
        local buildingInfo = GameInfo.Buildings[typeID]
        print("BuildingConstructed: " .. playerID .. ", " .. buildingInfo.BuildingType .. ", " .. plotX .. ", ".. plotY)
        
        if religionTypeCreated ~= -1 then
            if buildingInfo.BuildingType == "BUILDING_SHRINE" then
                for i = 1, 2 do
                    playerUnits:Create(GameInfo.Units["UNIT_MISSIONARY"].Index, plotX, plotY)
                end
                print("Spawned 2 UNIT_MISSIONARY at " .. plotX .. ", " .. plotY)
            elseif buildingInfo.BuildingType == "BUILDING_TEMPLE" then
                -- Spawn two apostles
                for i = 1, 2 do
                    playerUnits:Create(GameInfo.Units["UNIT_APOSTLE"].Index, plotX, plotY)
                end
                print("Spawned 2 UNIT_APOSTLE at " .. plotX .. ", " .. plotY)
            end
        end
    else
        local districtInfo = GameInfo.Districts[typeID]
        print("DistrictConstructed: " .. playerID .. ", " .. districtInfo.DistrictType .. ", " .. plotX .. ", ".. plotY)

        if districtInfo.DistrictType == "DISTRICT_HOLY_SITE" then
            local prophetClassIndex = GameInfo.GreatPersonClasses["GREAT_PERSON_CLASS_PROPHET"].Index
            local greatPeoplePoints = player:GetGreatPeoplePoints()
            greatPeoplePoints:ChangePointsTotal(prophetClassIndex, 10)
            print("Added 8 Great Prophet points to player " .. playerID)
        end
    end
end

-- Event handler for district construction
function OnDistrictConstructed(playerID, districtID, iX, iY)
    OnConstructionCompleted(playerID, districtID, iX, iY, false)
end

-- Event handler for building construction
function OnBuildingConstructed(playerID, cityID, buildingID, plotID, bOriginalConstruction)
    local plot = Map.GetPlotByIndex(plotID)
    if plot then
        local plotX = plot:GetX()
        local plotY = plot:GetY()
        OnConstructionCompleted(playerID, buildingID, plotX, plotY, true)
    else
        print("Error: Invalid plotID " .. plotID)
    end
end

-- Function to handle unit deaths
function OnUnitKilledInCombat(killedPlayerID, unitID, killingPlayerID, killingUnitID)
    local killingPlayer = Players[killingPlayerID]
    local killedPlayer = Players[killedPlayerID]

    -- Check if the killing player is an AI and the killed player is valid
    if killingPlayer and not killingPlayer:IsHuman() and killingPlayer:IsMajor() and killedPlayer then
        -- Get the defeated unit's combat strength
        local pKilledUnit = UnitManager.GetUnit(killedPlayerID, unitID)
        if pKilledUnit then
            local unitType = pKilledUnit:GetType()
            local unitInfo = GameInfo.Units[unitType]
            if unitInfo and unitInfo.Combat ~= nil then
                local combatStrength = unitInfo.Combat
                local goldReward = combatStrength * 7

                -- Add gold to the AI player's treasury
                local playerTreasury = killingPlayer:GetTreasury()
                playerTreasury:ChangeGoldBalance(goldReward)

                print("AI Player " .. killingPlayerID .. " received " .. goldReward .. " gold for defeating unit of combat strength " .. combatStrength)
            else
                print("UnitInfo or CombatStrength not found for unit type " .. unitType)
            end
        else
            print("Killed unit not found")
        end
    end
end

function OnUnitMoved(playerID, unitID, x, y)
    local pPlayer = Players[playerID]
    if pPlayer:IsHuman() then
        local pUnit = pPlayer:GetUnits():FindID(unitID)
        if pUnit then
            local pPlayerVisibility = PlayersVisibility[playerID]
            local currentTurn = Game.GetCurrentGameTurn()

            -- Loop over all AI players
            for _, otherPlayerID in ipairs(PlayerManager.GetAliveIDs()) do
                local otherPlayer = Players[otherPlayerID]
                if not otherPlayer:IsHuman() and otherPlayer:IsMajor() then
                    if not otherPlayer:IsHuman() then
                        -- Loop over all units of the AI player
                        for _, otherUnit in otherPlayer:GetUnits():Members() do
                            local unitX, unitY = otherUnit:GetX(), otherUnit:GetY()
                            if pPlayerVisibility:IsVisible(unitX, unitY) then
                                -- Set the property
                                -- print("Unit: " .. otherUnit:GetName() .. "(" .. otherUnit:GetX() .. "," .. otherUnit:GetY() .. ") set last_visible_in_turn to " .. currentTurn)
                                otherUnit:SetProperty("last_visible_in_turn", currentTurn)
                            end
                        end
                    end
                end
            end
        end
    end
end


-- Register the function to be called at the beginning of each turn
Events.PlayerTurnActivated.Add(OnPlayerTurnActivated)
GameEvents.OnDistrictConstructed.Add(OnDistrictConstructed)
GameEvents.BuildingConstructed.Add(OnBuildingConstructed)
Events.UnitKilledInCombat.Add(OnUnitKilledInCombat)
Events.UnitMoved.Add(OnUnitMoved)
print("loaded apply.lua")



--[[            -- Apply Spread Distance Modifier
local spreadDistanceModifier = GameInfo.Modifiers[MODIFIER_ID_SPREAD_DISTANCE]
if spreadDistanceModifier then
    pPlayer:GetReligion():AddModifier(MODIFIER_ID_SPREAD_DISTANCE)
end

-- Apply Spread Strength Modifier
local spreadStrengthModifier = GameInfo.Modifiers[MODIFIER_ID_SPREAD_STRENGTH]
if spreadStrengthModifier then
    pPlayer:GetReligion():AddModifier(MODIFIER_ID_SPREAD_STRENGTH)
end ]]